#!/usr/bin/env python3

import sys
import os
import os.path
import errno
import subprocess
import argparse
import datetime as dt
import logging
import tempfile
import glob
import shutil

from Config import *
from FileSystemUtils import *

_VCID="$Id: CompressCompletedRuns.py,v a00ea4a255f1 2013-02-17 19:32:25Z garrett $"


def compress_completed(sample_root_dir,
                analysis_root_dir,
                problem_name,
                log_level,
                time_limit):

    start_time = dt.datetime.now()
    running_time = dt.timedelta(seconds=time_limit)
    end_time = start_time + running_time

    logging.info("Starting at " + start_time.isoformat(' '))
    logging.info("Stop by " + end_time.isoformat(' '))

    sample_collection_dir = os.path.join(sample_root_dir, problem_name)
    analysis_collection_dir = os.path.join(analysis_root_dir, problem_name)

    sample_dirs = os.listdir(sample_collection_dir)
    sample_dirs.sort()

    for d in sample_dirs:
        # check time (1)
        remaining_time = end_time - dt.datetime.now()
        if remaining_time < dt.timedelta(seconds=5*60):
            logging.info("Out of time, stopping")
            break

        logging.info("Working on {}".format(d))

        report_file_path = os.path.join(analysis_collection_dir, 
                                        "{}.json".format(d))

        sample_dir = os.path.join(sample_collection_dir, d)

        # look for a report file and skip this sample if it's not there
        if not os.path.exists(report_file_path):
            logging.info("Sample  has no report; skipping")
            continue

        # If the sample is not marked complete, skip it
        complete_file_path = os.path.join(sample_dir, "complete")
        if not os.path.exists(complete_file_path):
            logging.info("Sample is not marked complete; skipping")
            continue

        # If we can get a lock, fine, otherwise, go on to the next sample:
        with DirLock(sample_dir) as lock:
            if lock is None:
                logging.info(
                    "Unable to lock {}, trying another"
                    .format(sample_dir))
                continue

            logging.info("Acquired lock for {}".format(sample_dir))

            log_file_path = os.path.join(sample_dir, "Log.txt")
            
            with tempfile.TemporaryDirectory(prefix="gmitchener_",
                                             dir=WORKSPACE_ROOT_DIR) as work_dir, RedirectStd(log_file_path):
                
                logging.info("***Begin Work {}***".format(dt.datetime.now()))

                # Copy stuff out of the archive to a local scratch folder
                # (because compression on the archive file system is SLOW
                # and it doesn't admit mmap'ed files):

                sample_work_dir = os.path.join(work_dir, "Sample")
                shutil.copytree(sample_dir, sample_work_dir)


                # Compress key databases
                to_compress = ["agents", "genomes", "generations"]

                try:
                    for file_name in to_compress:
                        compressed_file_name = file_name + ".bz2"
                        working_file_path = os.path.join(sample_work_dir, file_name)
                        compressed_file_path = os.path.join(sample_work_dir, compressed_file_name)
                        archive_file_path = os.path.join(sample_dir, file_name)
                        archive_compressed_file_path = os.path.join(sample_dir, compressed_file_name)
                        
                        if os.path.exists(archive_compressed_file_path):
                            logging.info("File already compressed: {}".format(archive_file_path))
                        elif os.path.exists(working_file_path):
                            logging.info("Compressing {}".format(file_name))
                            compress(working_file_path)
                            shutil.copyfile(compressed_file_path, archive_compressed_file_path)
                            os.remove(archive_file_path)
                        else:
                            logging.error("File is missing: {}".format(working_file_path))
                            raise LostFileError(working_file_path)
                                                                       
                except Exception as e:
                    logging.error("Error during compression: {}".format(e))

                logging.info("***End Work {}***".format(dt.datetime.now()))

            # end of with temp dir
        # end of with lock
    # end of main loop

    logging.info("Stopping at " + dt.datetime.now().isoformat(' '))
    # end of function

# Because bool('False') = True !?
def boolean(s):
    if s in ['True', 'true', 'yes']:
        return True
    elif s in ['False', 'false', 'no']:
        return False
    else:
        raise ValueError(s)

p = argparse.ArgumentParser(
    description="Compress database files in completed samples that have been analyzed",
    fromfile_prefix_chars="@")
p.add_argument('problem_name', help="Which problem to run")
p.add_argument('--sample-root-dir',
               help="Directory housing sample directories",
               default=SAMPLE_ROOT_DIR)
p.add_argument('--analysis-root-dir',
               help="Directory to put analysis files",
               default=ANALYSIS_ROOT_DIR)
p.add_argument('--time-limit',
               help="Maximum total time in seconds to spend",
               type=int,
               default=30*60)
p.add_argument('--log-level',
               help="Show messages at this level or above",
               choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'),
               default='ERROR')

def main():
    args = p.parse_args()
    log_level = getattr(logging, args.log_level.upper())
    logging.basicConfig(level=log_level, stream=sys.stderr)
    logging.debug("Args: {}".format(args))
    try:
        compress_completed(**vars(args))
    finally:
        logging.shutdown()

if __name__ == '__main__':
    main()
