#!/usr/bin/env python3

import sys
import os
import os.path
import errno
import subprocess
import argparse
import datetime as dt
import logging
import tempfile
import glob
import shutil

from Config import *
from FileSystemUtils import *

_VCID="$Id: DeleteCompletedRuns.py,v 80a13caf0407 2013-02-22 19:08:46Z garrett $"


def delete_completed(sample_root_dir,
                     analysis_root_dir,
                     problem_name,
                     log_level,
                     time_limit,
                     yes_do_it):

    start_time = dt.datetime.now()
    running_time = dt.timedelta(seconds=time_limit)
    end_time = start_time + running_time

    logging.info("Starting at " + start_time.isoformat(' '))
    logging.info("Stop by " + end_time.isoformat(' '))

    sample_collection_dir = os.path.join(sample_root_dir, problem_name)
    analysis_collection_dir = os.path.join(analysis_root_dir, problem_name)

    sample_dirs = os.listdir(sample_collection_dir)
    sample_dirs.sort()

    for d in sample_dirs:
        # check time (1)
        remaining_time = end_time - dt.datetime.now()
        if remaining_time < dt.timedelta(seconds=5*60):
            logging.info("Out of time, stopping")
            break

        logging.info("Working on {}".format(d))

        report_file_path = os.path.join(analysis_collection_dir, 
                                        "{}.json".format(d))

        sample_dir = os.path.join(sample_collection_dir, d)

        # look for a report file and skip this sample if it's not there
        if not os.path.exists(report_file_path):
            logging.info("Sample  has no report; skipping")
            continue

        # If the sample is not marked complete, skip it
        # complete_file_path = os.path.join(sample_dir, "complete")
        # if not os.path.exists(complete_file_path):
        #    logging.info("Sample is not marked complete; skipping")
        #    continue

        # If we can get a lock, fine, otherwise, go on to the next sample:
        with DirLock(sample_dir) as lock:
            if lock is None:
                logging.info(
                    "Unable to lock {}, trying another"
                    .format(sample_dir))
                continue

            logging.info("Acquired lock for {}".format(sample_dir))

            log_file_path = os.path.join(sample_dir, "Log.txt")
            
            logging.info("***Begin Work {}***".format(dt.datetime.now()))

            # Delete these large files
            to_delete = ["agents", "genomes", "generations"]

            def maybe_remove(file_path):
                if os.path.exists(file_path):
                    if yes_do_it:
                        logging.info("Removing {}".format(file_path))
                        os.remove(file_path)
                    else:
                        logging.info("Would remove {}".format(file_path))
                    

            try:
                for file_name in to_delete:
                    working_file_path = os.path.join(sample_dir, file_name)
                    compressed_file_name = file_name + ".bz2"
                    compressed_file_path = os.path.join(sample_dir,
                                                        compressed_file_name)
                        
                    maybe_remove(compressed_file_path)
                    maybe_remove(working_file_path)


            except Exception as e:
                logging.error("Error during deletion: {}".format(e))

            logging.info("***End Work {}***".format(dt.datetime.now()))

        # end of with lock
    # end of main loop

    logging.info("Stopping at " + dt.datetime.now().isoformat(' '))
    # end of function

# Because bool('False') = True !?
def boolean(s):
    if s in ['True', 'true', 'yes']:
        return True
    elif s in ['False', 'false', 'no']:
        return False
    else:
        raise ValueError(s)

p = argparse.ArgumentParser(
    description="Compress database files in completed samples that have been analyzed",
    fromfile_prefix_chars="@")
p.add_argument('problem_name', help="Which problem to run")
p.add_argument('--yes-do-it',
               help="Whether to actually do the work (since this is so destructive)",
               type=boolean,
               default=False)
p.add_argument('--sample-root-dir',
               help="Directory housing sample directories",
               default=SMALL_SAMPLE_ROOT_DIR)
p.add_argument('--analysis-root-dir',
               help="Directory of analysis files",
               default=ANALYSIS_ROOT_DIR)
p.add_argument('--time-limit',
               help="Maximum total time in seconds to spend",
               type=int,
               default=30*60)
p.add_argument('--log-level',
               help="Show messages at this level or above",
               choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'),
               default='ERROR')

def main():
    args = p.parse_args()
    log_level = getattr(logging, args.log_level.upper())
    logging.basicConfig(level=log_level, stream=sys.stderr)
    logging.debug("Args: {}".format(args))
    try:
        delete_completed(**vars(args))
    finally:
        logging.shutdown()

if __name__ == '__main__':
    main()
