# Utilities for dealing with various files and directories

import sys
import os
import os.path
import errno
import logging
import subprocess
import shutil

from Config import *

class DatabaseError(Exception):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return str(self.path)

class CorruptDatabaseError(DatabaseError):
    pass

class FileDoesNotExistError(DatabaseError):
    pass

def remove_file_safe(path):
    try:
        if os.path.exists(path):
            os.remove(path)
    except OSError:
        pass

def verify_exists(path):
    logging.debug("Checking existence of " + path)
    if os.path.exists(path):
        return True
    elif os.path.exists(path + ".bz2"):
        logging.info("Decompressing " + path + ".bz2")
        uncompress(path + ".bz2")
    else:
        raise FileDoesNotExistError(path)

def verify_db(path):
    verify_exists(path)
    logging.debug("Checking integrity of " + path)
    e_code = subprocess.call(["db_verify", path])
    check = e_code == 0
    if not check:
        raise CorruptDatabaseError(path) 
    else:
        return True

class AlreadyLockedException(Exception):
    pass

def lock_file(lock_file_path):
    try:
        return os.open(lock_file_path, os.O_CREAT | os.O_EXCL | os.O_RDWR)
    except OSError as e:
        # If we just couldn't get the lock, no message, but
        # if something else is going on, re-raise the
        # exception
        if e.errno == errno.EEXIST:
            raise AlreadyLockedException(lock_file_path)
        else:
            raise
    
class DirLock:
    def __init__(self, dir_path):
        self.dir_path = dir_path
        self.lock_file_path = os.path.join(dir_path, "lock")

    def __enter__(self):
        try:
            self.fd = lock_file(self.lock_file_path)
            # Make a stream wrapper
            # closefd=False means don't close underlying fd
            # when stream wrapper is closed.
            output = open(self.fd, "wt", closefd=False)
            output.write("Command line:\n")
            for a in sys.argv:
                output.write(a)
                output.write("\n")
            output.write("\nEnvironment:\n".format(os.environ))
            for (k, v) in os.environ.items():
                output.write('{}: {}\n'.format(k, v))
            output.close()
            return self
        except AlreadyLockedException:
            self.fd = None
            return None
        
    def __exit__(self, exc_type, exc_value, traceback):
        if self.fd is not None:
            os.close(self.fd)
            remove_file_safe(self.lock_file_path)
        # Propagate any exception
        return False

# Resource manager for appending stdin and stdout to a log file.
class RedirectStd:

    def __init__(self, log_file_path):
        self.log_file_path = log_file_path
        self.log_file = None
        self.stdout = None
        self.stderr = None

    def __enter__(self):
        self.stdout = os.dup(sys.stdout.fileno())
        self.stderr = os.dup(sys.stderr.fileno())
        self.log_file = open(self.log_file_path, "at")
        os.dup2(self.log_file.fileno(),
                sys.stdout.fileno())
        os.dup2(self.log_file.fileno(),
                sys.stderr.fileno())

    def __exit__(self, exc_type, exc_value, traceback):
        os.dup2(self.stdout, sys.stdout.fileno())
        os.dup2(self.stderr, sys.stderr.fileno())
        self.log_file.close()

def compress(file_path):
    verbosity = ['-v'] if logging.getLogger().isEnabledFor(logging.DEBUG) else []
    subprocess.check_call(
        ["/usr/bin/bzip2"] + verbosity + [file_path])

def uncompress(file_path, keep=True):
    verbosity = ['-v'] if logging.getLogger().isEnabledFor(logging.DEBUG) else []
    subprocess.check_call(
        ["/usr/bin/bzip2", "-d"]
        + verbosity
        + (["-k"] if keep else [])
        + [file_path])

def replace_file(source_file_path, dest_dir):
    if not os.path.exists(source_file_path):
        logging.debug("Source file does not exist: {}".format(source_file_path))
        return

    (source_dir, just_file) = os.path.split(source_file_path)
    replace_file_path = os.path.join(dest_dir, just_file)
    logging.debug("Copying {} to {}"
                  .format(source_file_path,
                          replace_file_path))
    shutil.copyfile(source_file_path, replace_file_path)


class LostFileError(Exception):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "File is missing: {}".format(self.path)
