#!/usr/bin/env python3

import sys
import os
import os.path
import errno
import subprocess
import argparse
import datetime as dt
import logging
import tempfile
import glob
import shutil
import stat

from Config import *
from FileSystemUtils import *

_VCID="$Id: RunMany.py,v 0f364b5c3ac1 2013-03-03 21:50:00Z garrett $"


def run_samples(sample_root_dir,
                analysis_root_dir,
                problem_name,
                log_level,
                engine_args,
                sample_min, sample_max,
                time_limit,
                keep_dead, 
                trash_skips):

    start_time = dt.datetime.now()
    running_time = dt.timedelta(seconds=time_limit)
    end_time = start_time + running_time

    logging.info("Starting at " + start_time.isoformat(' '))
    logging.info("Stop by " + end_time.isoformat(' '))

    if not os.path.exists(sample_root_dir):
        logging.error("Directory does not exist: {}".format(sample_root_dir))
        return

    sample_collection_dir = os.path.join(sample_root_dir, problem_name)
    analysis_collection_dir = os.path.join(analysis_root_dir, problem_name)

    # If we're just starting, the directory might not exist;
    # These shouldn't ever cause an error, but they do for some
    # reason; some garbage with the cluster file system.
    try:
        os.makedirs(sample_collection_dir, exist_ok=True)
        os.makedirs(analysis_collection_dir, exist_ok=True)
    except OSError as e:
        logging.warning("Crazy makedirs error:", exc_info=e)

    # We stop as soon as possible if the constraint is for a maximum rating
    max_rating_goal = any(["max-rating" in arg for arg in engine_args])

    # Main loop
    m = sample_min
    while m <= sample_max:
        logging.debug("Checking sample #{} of {}".format(m, problem_name))

        # check time (1)
        remaining_time = end_time - dt.datetime.now()
        if remaining_time < dt.timedelta(seconds=5*60):
            logging.info("Out of time, stopping")
            break

        report_file_path = os.path.join(analysis_collection_dir, 
                                        "{:06}.json".format(m))

        # If the goal is a maximum rating, we do the following
        # checks and maybe skip this sample:

        if max_rating_goal:
            # look for a report file and skip this sample if it's there
            if os.path.exists(report_file_path):
                logging.info("Sample {:06} has a report; skipping".format(m))
                m += 1
                continue

        sample_dir = os.path.join(sample_collection_dir, "{:06}".format(m))

        # Create the directory if it's not there already;
        # there's a race condition here, and exist_ok=True doesn't mask
        # all exceptions
        try:
            os.makedirs(sample_dir, exist_ok=True)
        except OSError as e:
            logging.warning("Error reported while making directory {}:".format(sample_dir), exc_info=e)
            logging.warning("Ignoring, fingers crossed")

        # If we can get a lock, fine, otherwise, go on to the next m:
        with DirLock(sample_dir) as lock:
            if lock is None:
                logging.debug(
                    "Unable to lock {}, trying another"
                    .format(sample_dir))
                m += 1
                continue

            logging.info("Acquired lock for {}; beginning work".format(sample_dir))
            
            # Some runs are just plain ornery and until I figure out
            # what's going wrong, I'm going to mark them to be
            # skipped:

            skip_file_path = os.path.join(sample_dir, "skip")
            if os.path.exists(skip_file_path):
                if trash_skips:
                    logging.info("Sample {:06} marked 'skip'; moving to trash".format(m))
                    try:
                        temp_dir = tempfile.mkdtemp('', "trash-{:06}-".format(m), sample_collection_dir)
                        os.renames(sample_dir, temp_dir)
                        remove_file_safe(report_file_path)
                    except OSError as e:
                        # just in case
                        logging.warning("While trashing skipped sample {:06}, "
                                        "ignoring exception:"
                                        .format(m), exc_info=e)
                else:
                    logging.info("Sample {:06} marked 'skip'; skipping".format(m))
                # Either way, we've lost the lock, continue with the next sample
                m += 1
                continue

            # Try to get information about the directory
            try:
                extra_info = subprocess.check_output(['ls', '-al', sample_dir], 
                                                     universal_newlines=True)
                logging.info("Contents of {}:\n{}\n".format(sample_dir, extra_info))
            except Exception as e:
                logging.error("Something wrong with sample dir {}".format(sample_dir))
                logging.error("Going on to next sample")
                m += 1
                continue

            log_file_path = os.path.join(sample_dir, "Log.txt")

            with tempfile.TemporaryDirectory(
                prefix="gmitchener_",
                dir=WORKSPACE_ROOT_DIR) as work_dir, \
                RedirectStd(log_file_path):
                
                logging.info("***Begin Work {} on sample {:06}***".format(dt.datetime.now(), m))

                # Copy stuff out of the archive to a local scratch folder
                # (because compression on the archive file system is SLOW
                # and it doesn't admit mmap'ed files):

                sample_work_dir = os.path.join(work_dir, "Sample")
                    
                try:
                    shutil.copytree(sample_dir, sample_work_dir)
                    logging.info("Finished copying sample to local scratch")
                except Exception as e:
                    # Something went wrong; gather extra
                    # info and re-raise
                    logging.error("Something unexpected went wrong checking out sample {:06}:"
                                  .format(m), exc_info=e)
                    logging.error("Marking {:06} to be skipped".format(m))
                    with open(skip_file_path, "wt"):
                        pass
                    continue
                        
                # Keep track of whether anything needs to be copied back to the 
                # main source dir for this sample:

                work_dir_dirty = False

                problem_name_file_path = os.path.join(sample_work_dir, "problemName.txt")
                random_seed_file_path = os.path.join(sample_work_dir, "randomSeeds")
                meta_file_path = os.path.join(sample_work_dir, "meta")
                agents_file_path = os.path.join(sample_work_dir, "agents")
                genomes_file_path = os.path.join(sample_work_dir, "genomes")
                landmark_file_path = os.path.join(sample_work_dir, "landmarks")
                generations_file_path = os.path.join(sample_work_dir, "generations")
                complete_file_path = os.path.join(sample_work_dir, "complete")

                
                # If the problem name file doesn't exist, then the
                # sample directory has to be initialized fresh:

                try:
                    verify_exists(problem_name_file_path)
                except FileDoesNotExistError as e:
                    logging.info("Initializing")
                    remove_file_safe(complete_file_path)
                    work_dir_dirty = True
                    remove_file_safe(report_file_path)
                    subprocess.call(
                        [ENGINE, "initialize", sample_work_dir, problem_name,
                         '--keep-dead=' + str(keep_dead)])

                # Remove __db files so database verification can proceed *shrug*
                for f in glob.glob(sample_work_dir + '/__db.[0-9][0-9][0-9]'):
                    logging.debug("Removing DB environment file " + f)
                    remove_file_safe(f)

                # If the random seed file is corrupt, we have to
                # trash the run and restart it fresh:

                try:
                    verify_db(random_seed_file_path)
                except Exception as e:
                    logging.error("Marking sample {:06} to be skipped".format(m),
                                  exc_info=e)
                    with open(skip_file_path, "wt"):
                        pass
                    # Restart the loop
                    continue

                # Assuming the random seed file is okay, check the
                # essential database files.  If these are damaged,
                # the run can be reinitialized and re-run

                try:
                    verify_db(meta_file_path)
                    verify_db(agents_file_path)
                    verify_db(genomes_file_path)

                except DatabaseError as e:
                    logging.warning(str(e))
                    logging.info("Reinitializing sample")
                    work_dir_dirty = True
                    remove_file_safe(report_file_path)
                    try:
                        result = subprocess.check_output(
                            [ENGINE, "reinitialize", sample_work_dir,
                             '--keep-dead=' + str(keep_dead)])
                    except subprocess.CalledProcessError as e:
                        logging.error("Engine reinitialize terminated with an error {}"
                                      .format(e.returncode), exc_info=e)
                        logging.error("Marking sample {:06} to be skipped".format(m))
                        with open(skip_file_path, "wt"):
                            pass
                        logging.error("Output was as follows:\n***\n{}\n***\n"
                                      .format(e.output))
                        # Restart the loop
                        continue

                # Assuming the essential database files are okay,
                # check the non-essentials.  These will be
                # automaticallly rebuilt by Engine if needed.
                    
                try:
                    verify_db(landmark_file_path)
                except DatabaseError as e:
                    remove_file_safe(landmark_file_path)

                try:
                    verify_db(generations_file_path)
                except DatabaseError as e:
                    remove_file_safe(generations_file_path)


                # If the run is incomplete, try to finish running it
                while not os.path.exists(complete_file_path):
                    engine_run_time = end_time - dt.datetime.now() - dt.timedelta(minutes=5)
                    if engine_run_time < dt.timedelta(minutes=5):
                        logging.info("Out of time, stopping before run")
                        break
                    # Still trying to track down memory problem:
                    # I thought maybe if I limited running time,
                    # it would cut down the risk of heap clutter,
                    # but that didn't solve the memory problem
                    if engine_run_time > dt.timedelta(minutes=30):
                        engine_run_time = dt.timedelta(minutes=30)
                    logging.info("Running")
                    engine_run_cmd = ([ENGINE, 
                                       "run", sample_work_dir,
                                       '--log-level=' + log_level,
                                       '--max-clock-time={:d}'.format(int(engine_run_time.total_seconds()))
                                       ]
                                      + engine_args)
                    logging.debug("Calling engine: {}".format(engine_run_cmd))
                    work_dir_dirty = True
                    remove_file_safe(report_file_path)
                    e_code = subprocess.call(engine_run_cmd)
                    if e_code == 0:
                        if max_rating_goal:
                            logging.info("Marking run complete")
                            with open(complete_file_path, "w"):
                                pass
                            break
                    elif e_code == 20:
                        logging.info("Engine run reports that it didn't finish: {}".format(e_code))
                    else:
                        logging.error("Engine run terminated with a serious error {}".format(e_code))
                        logging.error("Marking sample {:06} to be skipped".format(m))
                        with open(skip_file_path, "wt"):
                            pass
                        remove_file_safe(complete_file_path)
                        break

                # If work was done, send it back to the main archive
                if work_dir_dirty:
                    logging.info("Copying files back to original sample dir")
                    for fn in [problem_name_file_path,
                               random_seed_file_path,
                               meta_file_path,
                               agents_file_path,
                               genomes_file_path,
                               landmark_file_path,
                               generations_file_path,
                               # Leave for last:
                               complete_file_path]:
                        replace_file(fn, sample_dir)
                    logging.info("Finished copying files back to original sample dir")

                # If the run is marked complete, maybe leave a report
                if os.path.exists(complete_file_path):
                    # If there's no report, make one
                    if os.path.exists(report_file_path):
                        logging.warning("Report file exists but it shouldn't; not remaking {}"
                                        .format(report_file_path))
                    else:
                        logging.info("Making report file")
                        report_cmd = [ENGINE, "list", sample_work_dir,
                                      "--select=landmarks",
                                      "--format=long-json",
                                      "--with-mutation-ratings=True",
                                      "--log-level=" + log_level]
                        try:
                            # (the universal_newlines=True makes the result a string,
                            # suitable for the write command below.  *shrug*)
                            result = subprocess.check_output(report_cmd, universal_newlines=True)
                            with open(report_file_path, "wt") as report_file:
                                report_file.write(result)
                        except subprocess.CalledProcessError as e:
                            logging.error("Marking sample {:06} to be skipped".format(m), exc_info=e)
                            with open(skip_file_path, "wt"):
                                pass
                            logging.error("Engine list terminated with an error {}"
                                          .format(e.returncode))
                            logging.error("Output was as follows:\n***\n{}\n***\n"
                                          .format(e.output))
                    # Either way, go on to the next m
                    m += 1
                # end of if complete file exists

                # If no complete file was found, the loop returns
                # to the same m before going on.

                logging.info("***End Work {} on sample {:06}***".format(dt.datetime.now(), m))

            # end of with temp dir
        # end of with lock
    # end of main loop

    logging.info("Stopping at " + dt.datetime.now().isoformat(' '))
    # end of function

# Because bool('False') = True !?
def boolean(s):
    if s in ['True', 'true', 'yes']:
        return True
    elif s in ['False', 'false', 'no']:
        return False
    else:
        raise ValueError(s)

p = argparse.ArgumentParser(
    description="Run many UM evolution samples; arguments other than these are passed to Engine",
    fromfile_prefix_chars="@")
p.add_argument('problem_name', help="Which problem to run")
p.add_argument('--sample-root-dir',
               help="Directory housing sample directories",
               default=SAMPLE_ROOT_DIR)
p.add_argument('--analysis-root-dir',
               help="Directory to put analysis files",
               default=ANALYSIS_ROOT_DIR)
p.add_argument('--time-limit',
               help="Maximum total time in seconds to spend running samples",
               type=int,
               default=30*60)
p.add_argument('--sample-min',
               help="Minimum sample index to run to",
               type=int,
               default=1)
p.add_argument('--sample-max',
               help="Maximum sample index to run to",
               type=int,
               default=250)
p.add_argument('--keep-dead',
               help="Whether to keep genomes of dead agents around",
               type=boolean,
               default=True)
p.add_argument('--log-level',
               help="Show messages at this level or above",
               choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'),
               default='ERROR')
p.add_argument('--trash-skips',
               help="Whether to trash runs marked skip",
               type=boolean,
               default=True)

LOG_FORMAT='{levelname} {asctime} {filename}:{lineno}, {funcName}:\n>> {message}'

def main():
    (args, engine_args) = p.parse_known_args()
    log_level = getattr(logging, args.log_level.upper())
    logging.basicConfig(level=log_level, stream=sys.stderr, style='{', format=LOG_FORMAT)
    logging.info("Hostname: {}".format(os.environ['HOSTNAME']))
    logging.info("Args: {}".format(args))
    logging.info("Engine args: {}".format(engine_args))

    try:
        run_samples(engine_args=engine_args, **vars(args))
    finally:
        logging.shutdown()

if __name__ == '__main__':
    main()
